from enum import Enum


class DropStyle(Enum):
    """
    Represents the three approaches to handling the dropping of objects
    """
    # No objects will be dropped, User sql scripts will take care of the dropping of objects
    NONE = 1

    # The schema will be dropped and recreated before models are generated. This method prevents any objects from
    # accumulating in the schema
    SCHEMA = 2

    # The table will be dropped before the model is generated
    TABLE = 3



class ConfigFormats(Enum):
    """
    Enumeration of the configuration formats supported by DataBeaver
    """
    # cfg/ini format
    INI = 'ini'

    # JSON format
    JSON = 'json'

    # TOML format
    TOML = 'toml'

    # YAML format
    YAML = 'yaml'


class ModelStatus(Enum):
    """
    Enumeration of the status of a given model (table)
    """
    # The model was successfully realized
    NOT_BUILT = "not built"

    # We could not successfully realize the model
    FAILED = "failed"

    # The model was successfully realized
    BUILT = "built"

    # We could not build the model due to dependent models that failed to be built
    SKIPPED = "skipped"


class ExecutionStatus(Enum):
    """
    Enumeration for the sql file execution statuses
    """

    # The file has not yet been executed
    NOT_EXECUTED = "not executed"

    # The file is now being executed
    RUNNING = "running"

    # Execution failed
    FAILED = "failed"

    # File was skipped due to upstream files failing
    SKIPPED = "skipped"

    # File was executed successfully
    SUCCEEDED = "succeeded"


class Systems(Enum):
    MYSQL = "MySQL"

    POSTGRESQL = "PostgreSQL"



import psycopg2
import psycopg2.extras
import logging

from .DataStorage import DataStorage
import sys


class PostgreSQL(DataStorage):
    """
    Provides PostgreSQL database functionality. Implements the DataStorage abstraction.
    """

    def __init__(self, host, database, user, password):
        """

        :param host: Host to connect to
        :param database: Database to connect
        :param user: Username to connect with
        :param password: Password to connect with
        """
        # Call DataStorage.__init__()
        super().__init__()

        # Used to hold the connection to postgres
        self.connection = None
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.type = self.STORAGE_TYPE_POSTGRESQL

        # Connect to postgres
        self.connect(self.host, self.database, self.user, self.password)

    ##################
    # Public Methods #
    ##################
    def connect(self, host, database, user, password):
        """
        Connect to the server. If a connection already exists do nothing.

        :return: True if a connection was created. Otherwise False.
        """

        # If we already have a connection then do nothing
        if self.connection is None:
            self.connection = psycopg2.connect(host=host,
                                               database=database,
                                               user=user,
                                               password=password)
            return True

        # We did not create a new connection during this call.
        # This does not mean we are not connected...
        return False

    def create_table(self, schema='', table_name=None, columns=None):
        """
        Description
            Creates a new table in the requested based on the column information supplied

        Responsible For
            1. Seeing if the table already exists in the given schema and raises an error as needed
            2. Creating the table as specified

        Arguments
            :param schema: Schema to write table into
            :param table_name: Name of the new table we want to create
            :param columns: Column Information for the table we want to create
            :return: True if a new table was created, False otherwise
        """
        logger = self.get_logger()

        # Check that the table_name was supplied
        if table_name is None:
            logger.warn("table_name must be defined")
            return

        # Format the schema to use in sql below
        if schema != '':
            schema = f"{schema}."

        # Format the column information in a PostgreSQL specific format
        column_list = ""
        for column in columns:
            if 'type' in column:
                data_type = column['type']
            else:
                data_type = "VARCHAR(255)"
            column_list += f"\n{column['name']} {data_type},"
        column_list = column_list[0:-1]

        try:
            cursor = self.connection.cursor()

            # create the table
            command = f"CREATE TABLE {schema}{table_name}\n(\n{column_list}\n)"
            logger.debug(command)
            cursor.execute(command)

            # close communication with the PostgreSQL database server
            cursor.close()

            # commit the changes
            self.connection.commit()

        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

            # Transaction has been aborted, we need to close the existing cursor and get
            cursor.close()

            # Restart the connection
            self.connection = None
            self.connect(self.host, self.database, self.user, self.password)

            # Get new cursor (not sure if I need this but for now)
            cursor = self.connection.cursor()
        finally:
            cursor.close()

    def delete_rows(self, table_name, keys_to_delete={}, schema="public", log_level=logging.DEBUG):
        """
        Delete Rows the given table

        :param table_name:
        :param keys_to_delete:
        :param schema: Schema the table resides in. Default is 'public'
        :param log_level: Logging in this method will be done at the specified level. Default debug
        :return:
        """
        logger = self.get_logger()

        # Delete all rows in the table
        where_clause = "WHERE 1=1\n"
        for key, values in keys_to_delete.items():
            new_values = []
            for val in values:
                new_values.append(f"'{val}'")

            new_values = ", ".join(new_values)
            where_clause += f"AND {str(key)} IN ({new_values})\n"

        delete_stmt = f"DELETE FROM {schema}.{table_name}\n{where_clause};\n"
        try:
            cursor = self.connection.cursor()

            # create the table
            logger.log(log_level, delete_stmt)
            cursor.execute(delete_stmt)

            rows_deleted = cursor.rowcount
            logger.log(log_level, f"{rows_deleted} rows were deleted from {table_name}")

        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)
        finally:
            cursor.close()

        ###############################################
        # Close the Cursor and commit the transaction #
        ###############################################
        try:
            # Commit the changes
            self.connection.commit()

            # Close communication with the PostgreSQL database server
            cursor.close()

        except(Exception, psycopg2.DatabaseError) as error:
            logger.error(error)
        finally:
            cursor.close()

    def drop_table(self, table_name=None, schema="public"):
        """
        Drops the table in the given schema
        :param table_name:
        :param schema:
        :return:
        """
        logger = self.get_logger()

        drop_stmt = f"DROP TABLE {schema}.{table_name};"

        try:
            cursor = self.connection.cursor()

            # create the table
            cursor.execute(drop_stmt)
            logger.debug(drop_stmt)

            # close communication with the PostgreSQL database server
            cursor.close()

            # commit the changes
            self.connection.commit()

        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)
        finally:
            cursor.close()

    def execute(self, sql, ignore_errors=False, log_level=logging.DEBUG):
        # Get a _logger
        logger = self.get_logger()

        try:
            # Get a cursor
            cursor = self.connection.cursor()

            # Execute the statement
            logger.log(log_level, sql)
            cursor.execute(sql)
        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(sql)
            logger.error(error)
            if ignore_errors is False:
                raise error
        finally:
            cursor.close()

        # Close the Cursor and commit the transaction
        try:
            # Commit the changes
            self.connection.commit()

            # Close communication with the PostgreSQL database server
            cursor.close()

        except(Exception, psycopg2.DatabaseError) as error:
            logger.error(error)
            if ignore_errors is False:
                raise error
        finally:
            cursor.close()

    def execute_sql_file(self, filename, substitution_variables={},
                    cursor_factory=psycopg2.extras.RealDictCursor, log_level=logging.DEBUG):
        """

        :param log_level: level that method logging will be performed at
        :param cursor_factory: Defaults to list of dictionaries, pass in None to get a list of tuples
        :param filename:
        :param substitution_variables:
        :return:
        """
        template_file_path = f"{sys.path[0]}/sql/{filename}"
        with open(template_file_path) as f:
            raw_sql = f.read()

        sql = raw_sql
        for key, value in substitution_variables.items():
            sql = self.replace_text(sql, key, value)

        self.logger.log(log_level, sql)
        return self.postgresql.fetchall(sql, cursor_factory)

    def fetchall(self, query, cursor_factory=psycopg2.extras.RealDictCursor, log_level=logging.DEBUG):
        """

        :param cursor_factory: Set to None if you want list of tuples returned
        :param query: SQL statement to be executed
        :param log_level: Level to Log non error statements as. Defaults to debug
        :return result_set: Results From The Execution of <query>
        """
        logger = self.get_logger()

        with self.connection.cursor(cursor_factory=cursor_factory) as cursor:
            # Execute the SELECT EXISTS query to see if the table is already there or not
            cursor.execute(query)
            logger.log(log_level, query)

            # Return True if the table exists, false otherwise
            return cursor.fetchall()

    def insert_multi_rows(self, table_name, headers=None, data=None):
        """
        Inserts all rows in a single INSERT VALUES(), VALUES() statement

        :param table_name: Name of the table the data will be inserted into
        :param headers:
        :param data:
        :return:
        """
        logger = self.get_logger()

        if headers is None or data is None or table_name is None:
            logger.warn("No data to insert")
            return

        # Create a cursor to use for our processing
        try:
            cursor = self.connection.cursor()
        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

        # Loop over every row of data and attempt to insert it
        headers_string = ",".join(headers)
        insert_stmt = f"INSERT INTO {table_name} ({headers_string})\n VALUES\n"
        for row in data:
            row_string = []
            for value in row:
                value = str(value).replace("'", "")
                value = f"'{value}'"
                value = value.replace("'None'", "NULL")
                row_string.append(value)

            row_string = ", ".join(row_string)
            insert_stmt += f"({row_string}),\n"

        insert_stmt = f"{insert_stmt[:-2]};"
        try:
            # create the table
            logger.info(insert_stmt)
            cursor.execute(insert_stmt)
        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

        try:
            # close communication with the PostgreSQL database server
            cursor.close()

            # commit the changes
            self.connection.commit()
        except(Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

    def insert_list_of_tuples(self, table_name=None, schema=None, headers=None, data=None):
        """
        Generates individual insert statements and executes them against the database

        :param table_name: string - Table name to be loaded
        :param schema: string - Schema to use, will default to public if no schema is provided
        :param headers: list of strings - Columns names for the matching row data
        :param data: list of tuples -  Data to be inserted.

        :return: True if a new row was created, False otherwise
        """
        logger = self.get_logger()
        # Make sure we got data to work with
        if headers is None or data is None or table_name is None:
            logger.warn("No data to insert")
            return

        # Create a cursor to use for our processing
        try:
            cursor = self.connection.cursor()
        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

        # Determine if we have a schema or not, and build sql accordingly
        schema_clause = ''
        if schema:
            schema_clause = f"{schema}."

        # Loop over every row of data and attempt to insert it
        for row in data:
            # Create the insert statement for this row
            headers_string = ",".join(headers)

            row_string = []
            for value in row:
                value = str(value).replace("'", "")
                value = f"'{value}'"
                value = value.replace("'None'", "NULL")
                row_string.append(value)

            row_string = ", ".join(row_string)
            insert_stmt = f"INSERT INTO {schema_clause}{table_name} ({headers_string})\n VALUES({row_string});"

            try:
                # create the table
                cursor.execute(insert_stmt)
                logger.debug(insert_stmt)
            except (Exception, psycopg2.DatabaseError) as error:
                logger.error(error)

        try:
            # close communication with the PostgreSQL database server
            cursor.close()

            # commit the changes
            self.connection.commit()
        except(Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

    def insert_list_of_dictionaries(self, table_name, rows, schema=None, log_level=logging.DEBUG):
        """
        Inserts rows into a table from a list of dictionaries

        :param table_name: String - Name of the Table to Inserted Into
        :param rows: list of dictionaries - Rows to be inserted
        :param log_level: Set the log level for level configurable log messages
        :param schema: Optional schema to insert data into
        :return:
        """
        logger = self.get_logger()

        # Determine if we have a schema or not, and build sql accordingly
        schema_clause = ''
        if schema:
            schema_clause = f"{schema}."

        # Create a cursor to use for our processing
        try:
            cursor = self.connection.cursor()
        except (Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

        # Clean the data and group each record by the columns it would insert (so we can run them as one statement)
        data_to_insert = {}
        for row in rows:
            columns = []
            values = []
            for key, value in row.items():
                columns.append(key)
                value = str(value).replace("'", "")
                value = f"'{value}'"
                values.append(value)

            values_string = ", ".join(values)
            columns_string = ", ".join(columns)
            if columns_string not in data_to_insert:
                data_to_insert[columns_string] = ""
            data_to_insert[columns_string] += f"({values_string}),\n"

        # Insert all records grouped by the columns they will insert
        for columns, insert_values in data_to_insert.items():
            insert_stmt = f"INSERT INTO {schema_clause}{table_name} ({columns})\n VALUES\n{insert_values}"
            insert_stmt = f"{insert_stmt[:-2]};"
            insert_stmt = insert_stmt.replace("'None'", "NULL")
            insert_stmt = insert_stmt.replace("'Null'", "NULL")
            insert_stmt = insert_stmt.replace("''", "NULL")
            try:
                logger.log(log_level, insert_stmt)
                cursor.execute(insert_stmt)
            except (Exception, psycopg2.DatabaseError) as error:
                logger.error(insert_stmt)
                logger.error(error)
                break

        # Close the cursor and commit
        try:
            # close communication with the PostgreSQL database server
            cursor.close()

            # commit the changes
            self.connection.commit()
        except(Exception, psycopg2.DatabaseError) as error:
            logger.error(error)

    def list_schema(self):
        """
        List all schema in the database
        :return list_of_schema: List of the names of all schema
        """
        logger = self.get_logger()

        select_stmt = """
                      SELECT schema_name
                      FROM   information_schema.schemata
                      """
        with self.connection.cursor() as cursor:
            # Execute the query
            cursor.execute(select_stmt)
            logger.debug(select_stmt)

            # Fetch the result set
            results = cursor.fetchall()

        return results

    def table_exists(self, table_name=None, schema='public'):
        """

        :param schema:
        :param table_name:
        :return:
        """
        logger = self.get_logger()

        select_stmt = f"""SELECT EXISTS(
                              SELECT FROM information_schema.tables
                              WHERE table_schema = '{schema}'
                              AND table_name = '{table_name}');"""

        with self.connection.cursor() as cursor:
            # Execute the SELECT EXISTS query to see if the table is already there or not
            cursor.execute(select_stmt)
            logger.debug(select_stmt)

            # Return True if the table exists, false otherwise
            return cursor.fetchone()[0]



import re
import tomli
import os.path
import os
import psycopg2
import multiprocessing
import signal
import yaml
from .constants import ConfigFormats, ExecutionStatus, ModelStatus, Systems
from configparser import ConfigParser
from .DataStorage import DataStorage
from .Object import Object
from .PostgreSQL import PostgreSQL
from .MySQL import MySQL

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources


def _process_file(args):
    """
    Main multi thread worker process.
    Responsible For:
    1. Executing the sql for a given file

    :param args: Tuple containing input parameters to this invocation of this call
    :return: Filename That Was Executed, Execution Status, Database Error
    """
    filename_to_execute = args[0]
    sql_to_execute = args[1]
    system = args[2]

    if system['type'] == Systems.POSTGRESQL.value:
        return _postgresql_execute(system, filename_to_execute, sql_to_execute)
    elif system['type'] == Systems.MYSQL.value:
        return _mysql_execute(system, filename_to_execute, sql_to_execute)


def _mysql_execute(system, filename, sql):
    """
    :param system: Dictionary containing system level configuration (host, database, user, password)
    :param filename:
    :param sql:

    :return:
    """
    mysql = MySQL(system['host'], system['database'], system['user'], system['password'])

    try:
        mysql.execute(sql)
    except (Exception, psycopg2.DatabaseError) as error:
        return filename, ExecutionStatus.FAILED.value, error

    return filename, ExecutionStatus.SUCCEEDED.value, None


def _postgresql_execute(system, filename, sql):
    """

    :param system: Dictionary containing system level configuration (host, database, user, password)
    :param filename: Filename that is to be executed
    :param sql: SQL that will be executed
    :return:
    """
    postgres = PostgreSQL(system['host'], system['database'], system['user'], system['password'])

    try:
        postgres.execute(sql)
    except (Exception, psycopg2.DatabaseError) as error:
        return filename, ExecutionStatus.FAILED.value, error

    return filename, ExecutionStatus.SUCCEEDED.value, None


class DataBeaver(Object):
    """
    Responsible For
    - Data Model Orchestration (the building of 1 or more data models)
    """
    FIELD_PROCESSES = "processes"
    FIELD_SYSTEM = "system"
    SECTION_DATABEAVER = "DataBeaver"
    DEFAULT_LOGGER = "DataBeaver"
    EXECUTION_SUCCEEDED = "Succeeded"
    EXECUTION_FINISHED_STATUSES = [ExecutionStatus.FAILED.value, ExecutionStatus.SUCCEEDED.value,
                                   ExecutionStatus.SKIPPED.value]

    def __init__(self, base_file_path, config_file):
        """
        Initialize the DataBeaver model orchestration tool

        :param base_file_path: Will be appended to all file paths to make them absolute paths
        :param config_file:
        """

        # Call Object.__init__()
        super().__init__()

        self._config_format = None
        self._base_file_path = base_file_path
        self._config_file = f"{self._base_file_path}/{config_file}"
        self._config = None

        # Configure logging and instantiate the _logger
        self._logger = self.get_logger()

        # Determine the config file format if we do not yet know
        extension = config_file.split('.')[-1].lower().strip()
        if extension == 'ini':
            self._config_format = ConfigFormats.INI
        elif extension == 'toml':
            self._config_format = ConfigFormats.TOML
        elif extension == 'yaml':
            self._config_format = ConfigFormats.YAML
        elif extension == 'json':
            self._config_format = ConfigFormats.JSON

        # Load configuration from the supplied configuration file
        if self._config_format is ConfigFormats.TOML:
            with open(self._config_file, "rb") as f:
                self._config = tomli.load(f)
                # self._logger.info(self._config)
        elif self._config_format is ConfigFormats.INI:
            self._config = {}
            config = ConfigParser()
            config.read(self._config_file)
            for section in config.sections():
                self._config[section] = {}
                for option in config.options(section):
                    self._config[section][option] = config.get(section, option)
                    # Check if we need to substitute an environment variable
                    option_value = config.get(section, option)
                    env_variable_name = None
                    if str(option_value).lower().startswith('env:'):
                        env_variable_name = option_value[4:]
                    elif str(option_value).lower().startswith('environment:'):
                        env_variable_name = option_value[12:]

                    # Get the environment variable's value if needed
                    if env_variable_name:
                        option_value = os.environ.get(env_variable_name)

                    self._config[section][option] = option_value
        elif self._config_format is ConfigFormats.YAML:
            with open(self._config_file, 'r') as stream:
                self._config = yaml.safe_load(stream)

        # Supply default values as needed
        if self.SECTION_DATABEAVER not in self._config:
            self._config[self.SECTION_DATABEAVER] = {}

        if self.FIELD_PROCESSES not in self._config[self.SECTION_DATABEAVER]:
            self._logger.warning("Default number of processes not specified. Defaulting to 1")
            self._config[self.SECTION_DATABEAVER][self.FIELD_PROCESSES] = 1

    def _parse_sql(self, model_info, file_info, schema):
        """

        :param model_info: Model level information
        :param file_info: File level information
        :return: file_info: File Info (with added dependencies based off model references and parsed sql)
        """
        # Check For Model References To Add To File Dependencies
        for file_name, file in file_info.items():
            # Get the model name from the full file path
            model_name = file_name[file_name.rfind('/') + 1:file_name.rfind('.')]
            model_name = model_name[:model_name.find('.')]

            # Add any dependencies that are introduced by model references
            ref_positions = re.finditer(r"ref\(", file['raw_sql'])
            for match in ref_positions:
                close_tag_pos = file['raw_sql'].find(')', match.end())
                referred_model_name = file['raw_sql'][match.end():close_tag_pos]
                referred_model = model_info[referred_model_name]
                if referred_model_name != model_name \
                        and referred_model['steps'][-1] not in file_info[file_name]['dependencies']:
                    file_info[file_name]['dependencies'].append(referred_model['steps'][-1])

            # Create the compiled sql that we will need for execution
            tag_matches = re.finditer(r"\{\{", file['raw_sql'])
            current_position = 0
            parsed_model_sql = ""
            for match in tag_matches:
                close_tag_pos = file['raw_sql'].find(r"}}", match.end())
                tag_contents = file['raw_sql'][match.end() + 1:close_tag_pos]
                parsed_model_sql += file['raw_sql'][current_position: match.start()]
                tag_parsed = False

                # Process {{ ref(<model>) }} tags
                function_pos = tag_contents.find('ref(')
                if function_pos > -1:
                    pos = tag_contents.find("(")
                    end_pos = tag_contents.find(")", pos + 1)
                    referred_model_name = tag_contents[pos + 1:end_pos]
                    parsed_model_sql += f" {schema}.{referred_model_name} "

                # Process {{ full_table_name }} tags
                table_name_pos = tag_contents.find('full_table_name')
                if table_name_pos > -1:
                    parsed_model_sql += f"{schema}.{model_name}"
                    tag_parsed = True

                # Process {{ table_name }} tags (and only table_name)
                table_name_pos = tag_contents.find('table_name')
                if table_name_pos > -1 and tag_parsed is False:
                    parsed_model_sql += f"{model_name}"

                # Process {{ model_name }} tags
                table_name_pos = tag_contents.find('model_name')
                if table_name_pos > -1:
                    parsed_model_sql += f"{model_name}"

                # Set the current position where the tag was closed
                current_position = close_tag_pos + 2

            # Store the parsed sql for each file
            parsed_model_sql += file['raw_sql'][current_position:]
            file_info[file_name]['sql'] = parsed_model_sql

        return file_info

    def build(self, model=None):
        """
        Responsible for
        1. Determining model file dependencies
        2. Determining sql file dependencies
        3. Parsing sql files into executable code
        4. Executing the sql against the target system

        :return: (model_info, file_info)
        """

        # Look for models in the directories specified in the project file
        model_files = []
        for directory in self._config[model]['model_directories'].split(','):
            files = [f"{self._base_file_path}/{directory}/{file}" for file in os.listdir(f"{self._base_file_path}/{directory}")]
            model_files.extend(files)
        model_files = sorted(model_files)

        # Iterate over all the models and generate the compiled sql that we will run against the database
        # We will also generate the model dependencies
        model_info = {}
        file_info = {}
        for full_file_path in model_files:
            start_pos = full_file_path.rfind('/')
            model_name = full_file_path[start_pos + 1:]
            model_name = model_name[:model_name.find('.')]
            file_name = full_file_path[start_pos:]

            # Check if this is the first time we have seen this model
            if model_name not in model_info:
                model_info[model_name] = {'steps': [], 'current_step': 0, 'status': ModelStatus.NOT_BUILT.value}

            if full_file_path not in file_info:
                file_info[full_file_path] = {'file_name': file_name, 'model_name': model_name,
                                             'dependencies': [], 'status': ExecutionStatus.NOT_EXECUTED.value}

            # Add this model to the steps for the table
            model_info[model_name]['steps'].append(full_file_path)

            # Add any dependency based on having prior steps in the model to run
            if file_name.count('.') == 3:
                # Get the index number
                start_pos = file_name.find('.') + 1
                end_pos = file_name.find('.', start_pos, len(file_name))
                index_number = int(file_name[start_pos:end_pos])

                if index_number > 1:
                    previous_file_name = file_name.replace(f'.{index_number}.', f'.{index_number - 1}.')
                    previous_sql_name = full_file_path.replace(file_name, previous_file_name)
                    file_info[full_file_path]['dependencies'].append(previous_sql_name)
            else:
                self._logger.error(f"{full_file_path} can not be parsed and will be ignored.")
                continue

            # Load the sql out of the file
            with open(full_file_path, 'r') as model_file:
                raw_sql = model_file.read()
            file_info[full_file_path]['raw_sql'] = raw_sql

        file_info = self._parse_sql(model_info, file_info, self._config[model]['schema'])

        # Determine the number of processes to use, either by using a model specific configuration or default
        if self.FIELD_PROCESSES in self._config[model]:
            processes = int(self._config[model][self.FIELD_PROCESSES])
        else:
            processes = int(self._config[self.SECTION_DATABEAVER][self.FIELD_PROCESSES])

        for name, info in model_info.items():
            self._logger.info(f"{name}:{info}")

        # Get the system configuration
        system_config = self._config[self._config[model][self.FIELD_SYSTEM]]

        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        with multiprocessing.Pool(processes) as pool:
            signal.signal(signal.SIGINT, original_sigint_handler)

            # Loop over the models we will create until no more work can be done
            continue_processing = True
            loop_counter = 1

            while continue_processing:
                self._logger.info(f"Pass #{loop_counter}")
                loop_counter += 1

                # Determine all files we want to process in this pass
                files_to_process = []
                continue_processing = False
                for model_name in model_info.keys():
                    # Only process MODEL_NOT_BUILT models
                    if model_info[model_name]['status'] != ModelStatus.NOT_BUILT.value:
                        continue

                    # Get the file we will process for this model in this pass
                    current_step = model_info[model_name]['current_step']
                    filename_to_execute = model_info[model_name]['steps'][current_step]

                    # If all the dependencies have been satisfied, add the file to the list of files to be processed
                    if len(file_info[filename_to_execute]['dependencies']) == 0:
                        self._logger.info(f"{model_name} : {filename_to_execute} will be executed.")
                        files_to_process.append((filename_to_execute, file_info[filename_to_execute]['sql'],
                                                 system_config))

                # Execute all the sql statements for this pass in parallel
                results = pool.imap_unordered(_process_file, files_to_process)

                # Update file dependencies and model build status based on the results returned
                for (filename_executed, execution_status, db_error) in results:
                    # Get the model name from a file name
                    start_pos = filename_executed.rfind('/')
                    executed_model_name = filename_executed[start_pos + 1:]
                    executed_model_name = executed_model_name[:executed_model_name.find('.')]

                    # Update the file execution status for the file executed
                    file_info[filename_executed]['status'] = execution_status

                    # This file successfully executed and can be removed from any dependencies
                    if ExecutionStatus.SUCCEEDED.value == execution_status:
                        # Something succeeded so we need to keep processing
                        continue_processing = True
                        self._logger.info(f"{filename_executed}({model_info[executed_model_name]['current_step']}) - {execution_status}")

                        # Increment current_step
                        model_info[executed_model_name]['current_step'] = current_step + 1

                        # If we executed the last file in steps the model is built
                        if model_info[executed_model_name]['current_step'] == \
                                len(model_info[executed_model_name]['steps']):
                            model_info[executed_model_name]['status'] = ModelStatus.BUILT.value
                            self._logger.info(f"Setting status of {executed_model_name} to {ModelStatus.BUILT.value}")

                        for file_name, file in file_info.items():
                            if filename_executed in file['dependencies']:
                                file_info[file_name]['dependencies'].remove(filename_executed)

                    elif ExecutionStatus.FAILED.value == execution_status:
                        self._logger.info(f"{model_name} - {ModelStatus.FAILED.value}")
                        model_info[executed_model_name]['status'] = ModelStatus.FAILED.value

        return model_info, file_info

    def create_project(self, name, config_format=ConfigFormats.TOML):
        """
        Create a new empty
        :return:
        """
        self._logger.info('Creating new project')

        # Get the project name from the user and generate the directory name we will use
        directory_name = re.sub(' ', '_', name)
        directory_name = re.sub('[^A-Za-z0-9_]+', '', directory_name)
        self._logger.info(f"Project name is '{name}'")

        # Check if the directory already exists
        if os.path.isdir(directory_name):
            self._logger.error(f"{directory_name} already exists, operation can not be completed.")
            return

        # Make the top level project directory
        os.mkdir(directory_name)
        self._logger.info(f'Created {directory_name}')

        # Make the directory for configuration files
        config_directory = f"{directory_name}/system"
        os.mkdir(config_directory)
        self._logger.info(f'Created {config_directory}')

        # Get the data for the sample config file and set the file name for the sample config file
        config_sample = ''
        file_name = "unknown.config"
        if config_format is ConfigFormats.TOML:
            file_name = "databeaver.toml"
            config_sample = pkg_resources.read_text('databeaver.data', 'configSample.toml')
        elif config_format is ConfigFormats.YAML:
            file_name = "databeaver.yaml"
        elif config_format is ConfigFormats.INI:
            file_name = "databeaver.ini"
            config_sample = pkg_resources.read_text('databeaver.data', 'configSample.ini')
        elif config_format is ConfigFormats.JSON:
            file_name = "databeaver.json"

        # Write the config file to the file system
        file_path = f"{config_directory}/{file_name}"
        with open(file_path) as f:
            f.write(config_sample)

    @staticmethod
    def get_data_storage(storage_type, host, database, user, password):
        """
        Get realized DataStorage class. Currently only MySQL and PostgreSQL are supported.

        :param storage_type: What type of Data Storage do we want
        :param host: The host address for the data storage
        :param database: The database to use within the data storage
        :param user: User to connect to the DataStorage with
        :param password: Password to connect to the DataStorage with
        :return data_storage : PostgreSQL, MySQL, etc
        """
        data_storage = None
        if storage_type == DataStorage.STORAGE_TYPE_POSTGRESQL:
            data_storage = PostgreSQL(host, database, user, password)
        elif storage_type == DataStorage.STORAGE_TYPE_MYSQL:
            data_storage = MySQL(host, database, user, password)

        return data_storage



from .Object import Object
from .Table import Table


class Schema(Object):

    def __init__(self, name, data_storage):
        """
        :param system: System object
        :param name: Name of the schema

        Initialize a Schema object
        """
        # Call Object.__init__()
        super().__init__()

        # Set the name of the schema
        self.name = name

        # Set the system that this schema exists in
        self.data_storage = data_storage

    def list_tables(self):
        """

        :return:
        """
        return self.data_storage.list_tables(schema=self.name)

    def table_exists(self, table_object: Table = None):
        """
        Does this table exist in this schema

        :param table_object: Table we want to check for existence
        :return boolean: True if the table exists in this schema. False otherwise.
        """
        return self.data_storage.table_exists(schema=self.name, table_object=table_object)



from .Object import Object
import json


class Table(Object):
    """
    Abstract table class (not tied to any specific data storage)
    """
    def __init__(self, name, schema=None, definition_file=None, columns=None):
        self.name = name

        # Name of the schema that this table resides in
        self.schema = schema
        self.columns = []
        super().__init__()
        logger = self.get_logger()
        if definition_file is None and columns is None:
            logger.error(f"{self.name} is invalid. No definition file or fields provided.")

        # Load the column information
        if definition_file:
            with open(definition_file) as f:
                self.columns = json.load(f)
        else:
            self.columns = columns




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="databeaver",
    version="0.1.24",
    author="Fuzzy Bumblebee Software, LLC",
    author_email="david.orkin@fuzzybumblebee.org",
    description="Data Model Orchestration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://opensource.fuzzybumblebee.org/databeaver/latest",
    project_urls={
        "Bug Tracker": "https://github.com/FBB-David/databeaver/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: POSIX :: Linux"
    ],
    package_data={'databeaver': ['data/configSample.*']},
    install_requires=['tomli'],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.6",
    scripts=['src/databeaver/bin/beaver']
)






from .Object import Object
from .Schema import Schema
from .Table import Table


class DataStorage(Object):
    # Standard Data Types that all DataStorage classes must be able to handle
    DATA_TYPE_BOOLEAN = "BOOLEAN"
    DATA_TYPE_FLOAT = "FLOAT"
    DATA_TYPE_INTEGER = "INTEGER"
    DATA_TYPE_STRING = "STRING"
    DATA_TYPE_TEXT = "TEXT"
    DATA_TYPE_TIMESTAMP = "TIMESTAMP"

    STORAGE_TYPE_MYSQL = "MySQL"
    STORAGE_TYPE_POSTGRESQL = "PostgreSQL"
    STORAGE_TYPE_UNKNOWN = "UNKNOWN"

    def __init__(self):
        """
        Instantiate the base abstraction for all DataStorage classes
        """

        # Call Object.__init__()
        super().__init__()

        # Initial instantiation of class variables
        self.logger = self.get_logger()
        self.type = self.STORAGE_TYPE_UNKNOWN
        self.host = None
        self.database = None
        self.user = None
        self.password = None

    def create(self, schema=None, object_to_create=None):
        """
        Create a given Object in the database (withing the given schema
        :param schema:
        :param object_to_create: something we wish to create (Table, Tables, etc)
        :return:
        """
        if object_to_create is None:
            self.logger.warning(f"object_to_create is None. No work performed.")
            return

        # The caller requested a table be created
        if isinstance(object_to_create, Table):
            self.logger.info(f'Creating {schema}.{object_to_create.name}')
            self.create_table(schema, object_to_create.name, object_to_create.columns)

    def drop(self, schema=None, object_to_drop=None):
        """

        :param schema:
        :param object_to_drop:
        :return boolean: True if the table was dropped. False otherwise
        """
        if object_to_drop is None:
            self.logger.warning("No object to drop provided. No work done.")
            return False

        # The user requested a table be dropped
        if isinstance(object_to_drop, Table):
            self.logger.info(f"Dropping the table '{schema}.{object_to_drop.name}'")
            self.drop_table(table_name=object_to_drop.name, schema=schema)
            return True

        return False

    def get_schema(self, schema_name):
        """
        Get an instance of a Schema class

        :param schema_name:
        :return schema: Instantiated Schema object
        """
        schema = Schema(schema_name, data_storage=self)
        return schema

    def list_tables(self, schema=None):
        """
        List all tables in a given schema

        :param schema:
        :return:
        """
        self.list_tables(schema)

    def exists(self, schema=None, object_to_check = None):
        """
        Does a table exist in a given schema
        :return:
        """

        if isinstance(object_to_check, Table):
            return self.table_exists(schema=schema, table_name=object_to_check.name)

        return False


import collections
import logging
import re



class Object:
    """
    Base Object class for all objects in our system.

    Responsible For
    1. Logger Configuration (via get_logger)
    2. Conversion of data format (via create_dict, flatten, string_to_int, string_to_float, )
    3. Standard template substitution (via replace_text)
    4.

    """
    @staticmethod
    def get_logger():

        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=[
                logging.FileHandler("debug.log"),
                logging.StreamHandler()
            ]
        )

        return logging.getLogger()

    @staticmethod
    def create_dict(list_of_dict, keys, value_key):
        """
        Converts a list of dictionaries into a dictionary
        :param list_of_dict: List of dictionary values we want to convert
        :param keys: The first key is considered the top level key and all subsequent keys are used to insert values
        :param value_key: Key in each dictionary that will be used for the values
        :return:
        """
        result = {}
        for dict_item in list_of_dict:
            top_key_value = dict_item[keys[0]]
            if top_key_value not in result:
                result[top_key_value] = {}

            for key in keys[1:]:
                key_value = dict_item[key]
                result[top_key_value][key_value] = dict_item[value_key]

        return result

    def flatten(self, dictionary, parent_key='', sep='_'):
        items = []
        for k, v in dictionary.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, collections.MutableMapping):
                items.extend(self.flatten(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    @staticmethod
    def replace_text(text, key, value):
        """
        Substitutes a tag into a template.

        :return:
        """
        return text.replace(f"<!--[{key}]-->", str(value))

    @staticmethod
    def string_to_int(s):
        try:
            return int(s)
        except ValueError:
            return int('0')

    @staticmethod
    def string_to_float(s):
        try:
            return float(s)
        except ValueError:
            return float('0.00')

    @staticmethod
    def to_camel_case(text):
        """

        :param text: Text to convert to camel case
        :return:
        """
        # Replace the under bar character with a space character
        camel_case = re.sub("_", " ", text)
        camel_case = re.sub("-", " ", camel_case)

        # Capitalize each word we found in the class name
        camel_case = re.sub(r'\w+', lambda m: m.group(0).capitalize(), camel_case)

        # Remove the spaces to get the proper class name
        camel_case = re.sub(" ", "", camel_case)
        return camel_case
